;; Generated package description from dired-duplicates.el  -*- no-byte-compile: t -*-
(define-package "dired-duplicates" "0.2.0.20231109.135341" "Find duplicate files locally and remotely" '((emacs "27.1")) :commit "349eebbd8b3fc10fa2a4551426769d20744e2eb2" :authors '(("Harald Judt" . "h.judt@gmx.at")) :maintainer '("Harald Judt" . "h.judt@gmx.at") :keywords '("files") :url "https://codeberg.org/hjudt/dired-duplicates")
